// Copyright (C) 2016-2021  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package xneo
// transactional log of states

import (
	"sync"

	"lab.nexedi.com/kirr/neo/go/neo/proto"
)

// State represents state of a cluster.
type State struct {
	// XXX +transaction ID?

	NodeTab      *NodeTable         // nodes in the cluster
	PartTab      *PartitionTable    // on which nodes which data is located
	ClusterState proto.ClusterState // code for cluster state
}

// StateRO provides read-only access to a State.
// XXX -> StateSnapshot ?
type StateRO interface {
	NodeTab()      NodeTableRO
	PartTab()      PartitionTableRO
	ClusterState() proto.ClusterState
}

// StateLog is transactional journal of State changes.
type StateLog struct {
	mu   sync.Mutex
	head *State
}


// ReadHead returns state from head of the log.
func (l *StateLog) ReadHead() StateRO {
	l.mu.Lock()
	defer l.mu.Unlock()
	return l.head
}

// CommitUpdate creates new entry in the log with state updated by δf's changes.
func (l *StateLog) CommitUpdate(δf func(*State)) {
	l.mu.Lock()
	defer l.mu.Unlock()
	s := l.head.Clone()
	δf(s)
	l.head = s
}


// Clone creates new instance of State initialized with previous content.
func (s *State) Clone() *State {
	return &State{
		NodeTab:      s.NodeTab.Clone(),
		PartTab:      s.PartTab.Clone(),
		ClusterState: s.ClusterState,
	}
}
